@@ -568,3 +568,29 @@ class StringArray(StringMethods, ObjectArray):
         else:
             out = self._content[where]
             return self.__class__(out.starts, out.stops, out.content, self.encoding)
+
+    @property
+    def iscompact(self):
+        return self._content.iscompact
+
+    def compact(self):
+        return self.fromjagged(self._content.compact(), self.encoding)
+
+    def flatten(self):
+        return self.fromjagged(self._content.flatten(), self.encoding)
+
+    @awkward.util.bothmethod
+    def concatenate(isclassmethod, cls_or_self, arrays, axis=0):
+        if isclassmethod: 
+            cls = cls_or_self
+            if not all(isinstance(x, StringArray) for x in arrays):
+                raise TypeError("cannot concatenate non-StringArrays with StringArray.concatenate")
+        else:
+            self = cls_or_self
+            cls = self.__class__
+            if not isinstance(self, StringArray) or not all(isinstance(x, StringArray) for x in arrays):
+                raise TypeError("cannot concatenate non-StringArrays with StringArrays.concatenate")
+            arrays = (self,) + tuple(arrays)
+
+        jagged = self.JaggedArray.concatenate([x._content for x in arrays], axis=axis)
+        return self.fromjagged(jagged, self.encoding)
