@@ -30,6 +30,9 @@
 
 import json
 
+import numpy
+
+import awkward.array.base
 import awkward.array.chunked
 import awkward.array.indexed
 import awkward.array.jagged
@@ -40,6 +43,8 @@ import awkward.array.virtual
 import awkward.type
 import awkward.util
 
+################################################################################ type conversions
+
 def schema2type(schema):
     import pyarrow
 
@@ -123,7 +128,115 @@ def schema2type(schema):
 
     return out
 
-def view(obj, awkwardlib=None):
+################################################################################ value conversions
+
+def toarrow(obj):
+    import pyarrow
+
+    def recurse(obj, mask):
+        if isinstance(obj, numpy.ndarray):
+            return pyarrow.array(obj, mask=mask)
+
+        elif isinstance(obj, awkward.array.chunked.ChunkedArray):   # includes AppendableArray
+            raise TypeError("only top-level ChunkedArrays can be converted to Arrow (as RecordBatches)")
+
+        elif isinstance(obj, awkward.array.indexed.IndexedArray):
+            if mask is None:
+                return pyarrow.DictionaryArray.from_arrays(obj.index, recurse(obj.content, mask))
+            else:
+                return recurse(obj.content[obj.index], mask)
+
+        elif isinstance(obj, awkward.array.indexed.SparseArray):
+            return recurse(obj.dense, mask)
+
+        elif isinstance(obj, awkward.array.jagged.JaggedArray):
+            obj = obj.compact()
+            if mask is not None:
+                mask = obj._broadcast(mask).flatten()
+            return pyarrow.ListArray.from_arrays(obj.offsets, recurse(obj.content, mask))
+
+        elif isinstance(obj, awkward.array.masked.IndexedMaskedArray):
+            thismask = obj.boolmask(maskedwhen=True)
+            if mask is not None:
+                thismask = mask & thismask
+            if len(obj.content) == 0:
+                content = obj.numpy.empty(len(obj.mask), dtype=obj.DEFAULTTYPE)
+            else:
+                content = obj.content[obj.mask]
+            return recurse(content, thismask)
+
+        elif isinstance(obj, awkward.array.masked.MaskedArray):   # includes BitMaskedArray
+            thismask = obj.boolmask(maskedwhen=True)
+            if mask is not None:
+                thismask = mask & thismask
+            return recurse(obj.content, thismask)
+
+        elif isinstance(obj, awkward.array.objects.ObjectArray):
+            # throw away Python object interpretation, which Arrow can't handle while being multilingual
+            return recurse(obj.content, mask)
+
+        elif isinstance(obj, awkward.array.objects.StringArray):
+            # obj = obj.compact()
+            raise NotImplementedError("I don't know how to make an Arrow StringArray")
+
+            # I don't understand this
+            # pyarrow.StringArray.from_buffers(2, pyarrow.py_buffer(numpy.array([0, 5, 10])), pyarrow.py_buffer(b"helloHELLO"), offset=0)
+            # returns ["", "hello"]
+            # ???
+
+            # Also, be sure to check for the difference between strings and bytes!
+
+        elif isinstance(obj, awkward.array.table.Table):
+            return pyarrow.StructArray.from_arrays([recurse(x, mask) for x in obj.contents.values()], list(obj.contents))
+
+        elif isinstance(obj, awkward.array.union.UnionArray):
+            contents = []
+            for i, x in enumerate(obj.contents):
+                if mask is None:
+                    thismask = None
+                else:
+                    thistags = (obj.tags == i)
+                    thismask = obj.numpy.empty(len(x), dtype=obj.MASKTYPE)
+                    thismask[obj.index[thistags]] = mask[thistags]    # hmm... obj.index could have repeats; the Arrow mask in that case would not be well-defined...
+                contents.append(recurse(x, thismask))
+
+            return pyarrow.UnionArray.from_dense(pyarrow.array(obj.tags.astype(numpy.int8)), pyarrow.array(obj.index.astype(numpy.int32)), contents)
+
+        elif isinstance(obj, awkward.array.virtual.VirtualArray):
+            return recurse(obj.array, mask)
+
+        else:
+            raise TypeError("cannot convert type {0} to Arrow".format(type(obj)))
+
+    if isinstance(obj, awkward.array.chunked.ChunkedArray):   # includes AppendableArray
+        batches = []
+        for chunk in obj.chunks:
+            arr = toarrow(chunk)
+            if isinstance(arr, pyarrow.Table):
+                batches.extend(arr.to_batches())
+            else:
+                batches.append(pyarrow.RecordBatch.from_arrays([arr], [""]))
+        return pyarrow.Table.from_batches(batches)
+
+    elif isinstance(obj, awkward.array.masked.IndexedMaskedArray) and isinstance(obj.content, awkward.array.table.Table):
+        mask = obj.boolmask(maskedwhen=True)
+        if len(obj.content) == 0:
+            content = obj.numpy.empty(len(obj.mask), dtype=obj.DEFAULTTYPE)
+        else:
+            content = obj.content[obj.mask]
+        return pyarrow.Table.from_batches([pyarrow.RecordBatch.from_arrays([recurse(x, mask) for x in obj.content.contents.values()], list(obj.content.contents))])
+
+    elif isinstance(obj, awkward.array.masked.MaskedArray) and isinstance(obj.content, awkward.array.table.Table):   # includes BitMaskedArray
+        mask = obj.boolmask(maskedwhen=True)
+        return pyarrow.Table.from_batches([pyarrow.RecordBatch.from_arrays([recurse(x, mask) for x in obj.content.contents.values()], list(obj.content.contents))])
+
+    elif isinstance(obj, awkward.array.table.Table):
+        return pyarrow.Table.from_batches([pyarrow.RecordBatch.from_arrays([recurse(x, None) for x in obj.contents.values()], list(obj.contents))])
+
+    else:
+        return recurse(obj, None)
+
+def fromarrow(obj, awkwardlib=None):
     import pyarrow
     awkwardlib = awkward.util.awkwardlib(awkwardlib)
     ARROW_BITMASKTYPE = awkwardlib.numpy.uint8
@@ -133,7 +246,7 @@ def view(obj, awkwardlib=None):
 
     def popbuffers(tpe, buffers):
         if isinstance(tpe, pyarrow.lib.DictionaryType):
-            content = view(tpe.dictionary)
+            content = fromarrow(tpe.dictionary)
             index = popbuffers(tpe.index_type, buffers)
             if isinstance(index, awkwardlib.BitMaskedArray):
                 return awkwardlib.BitMaskedArray(index.mask, awkwardlib.IndexedArray(index.content, content), maskedwhen=index.maskedwhen, lsborder=index.lsborder)
@@ -155,7 +268,7 @@ def view(obj, awkwardlib=None):
         elif isinstance(tpe, pyarrow.lib.ListType):
             content = popbuffers(tpe.value_type, buffers)
             offsets = awkwardlib.numpy.frombuffer(buffers.pop(), dtype=ARROW_INDEXTYPE)
-            out = awkwardlib.JaggedArray.fromoffsets(offsets, content)
+            out = awkwardlib.JaggedArray.fromoffsets(offsets, content[:offsets[-1]])
             mask = buffers.pop()
             if mask is not None:
                 mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
@@ -170,6 +283,12 @@ def view(obj, awkwardlib=None):
             assert buffers.pop() is None
             tags = awkwardlib.numpy.frombuffer(buffers.pop(), dtype=ARROW_TAGTYPE)
             index = awkwardlib.numpy.arange(len(tags), dtype=ARROW_INDEXTYPE)
+            for i in range(len(contents)):
+                these = index[tags == i]
+                if len(these) == 0:
+                    contents[i] = contents[i][0:0]
+                else:
+                    contents[i] = contents[i][: these[-1] + 1]
             out = awkwardlib.UnionArray(tags, index, contents)
             mask = buffers.pop()
             if mask is not None:
@@ -184,6 +303,12 @@ def view(obj, awkwardlib=None):
                 contents.insert(0, popbuffers(tpe[i].type, buffers))
             index = awkwardlib.numpy.frombuffer(buffers.pop(), dtype=ARROW_INDEXTYPE)
             tags = awkwardlib.numpy.frombuffer(buffers.pop(), dtype=ARROW_TAGTYPE)
+            for i in range(len(contents)):
+                these = index[tags == i]
+                if len(these) == 0:
+                    contents[i] = contents[i][0:0]
+                else:
+                    contents[i] = contents[i][: these.max() + 1]
             out = awkwardlib.UnionArray(tags, index, contents)
             mask = buffers.pop()
             if mask is not None:
@@ -195,7 +320,7 @@ def view(obj, awkwardlib=None):
         elif tpe == pyarrow.string():
             content = awkwardlib.numpy.frombuffer(buffers.pop(), dtype=ARROW_CHARTYPE)
             offsets = awkwardlib.numpy.frombuffer(buffers.pop(), dtype=ARROW_INDEXTYPE)
-            out = awkwardlib.StringArray.fromoffsets(offsets, content, encoding="utf-8")
+            out = awkwardlib.StringArray.fromoffsets(offsets, content[:offsets[-1]], encoding="utf-8")
             mask = buffers.pop()
             if mask is not None:
                 mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
@@ -206,7 +331,7 @@ def view(obj, awkwardlib=None):
         elif tpe == pyarrow.binary():
             content = awkwardlib.numpy.frombuffer(buffers.pop(), dtype=ARROW_CHARTYPE)
             offsets = awkwardlib.numpy.frombuffer(buffers.pop(), dtype=ARROW_INDEXTYPE)
-            out = awkwardlib.StringArray.fromoffsets(offsets, content, encoding=None)
+            out = awkwardlib.StringArray.fromoffsets(offsets, content[:offsets[-1]], encoding=None)
             mask = buffers.pop()
             if mask is not None:
                 mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
@@ -247,19 +372,19 @@ def view(obj, awkwardlib=None):
         if len(chunks) == 1:
             return chunks[0]
         else:
-            return awkwardlib.ChunkedArray([view(x) for x in chunks], counts=[len(x) for x in chunks])
+            return awkwardlib.ChunkedArray([fromarrow(x) for x in chunks], counts=[len(x) for x in chunks])
 
     elif isinstance(obj, pyarrow.lib.RecordBatch):
         out = awkwardlib.Table()
         for n, x in zip(obj.schema.names, obj.columns):
-            out[n] = view(x)
+            out[n] = fromarrow(x)
         return out
 
     elif isinstance(obj, pyarrow.lib.Table):
         chunks = []
         counts = []
         for batch in obj.to_batches():
-            chunk = view(batch)
+            chunk = fromarrow(batch)
             if len(chunk) > 0:
                 chunks.append(chunk)
                 counts.append(len(chunk))
@@ -271,7 +396,82 @@ def view(obj, awkwardlib=None):
     else:
         raise NotImplementedError(type(obj))
 
-class ParquetFile(object):
+################################################################################ Parquet file handling
+
+def toparquet(obj, where, **options):
+    import pyarrow.parquet
+
+    options["where"] = where
+
+    def convert(obj, message):
+        if isinstance(obj, (awkward.array.base.AwkwardArray, numpy.ndarray)):
+            out = toarrow(obj)
+            if isinstance(out, pyarrow.Table):
+                return out
+            else:
+                return pyarrow.Table.from_batches([pyarrow.RecordBatch.from_arrays([out], [""])])
+        else:
+            raise TypeError(message)
+
+    if isinstance(obj, awkward.array.chunked.ChunkedArray):
+        obj = iter(obj.chunks)
+        try:
+            awkitem = next(obj)
+        except StopIteration:
+            raise ValueError("iterable is empty")
+
+        arritem = convert(awkitem, None)
+        if "schema" not in options:
+            options["schema"] = arritem.schema
+        writer = pyarrow.parquet.ParquetWriter(**options)
+        writer.write_table(arritem)
+
+        try:
+            while True:
+                try:
+                    awkitem = next(obj)
+                except StopIteration:
+                    break
+                else:
+                    writer.write_table(convert(awkitem, None))
+        finally:
+            writer.close()
+
+    elif isinstance(obj, (awkward.array.base.AwkwardArray, numpy.ndarray)):
+        arritem = convert(obj, None)
+        options["schema"] = arritem.schema
+        writer = pyarrow.parquet.ParquetWriter(**options)
+        writer.write_table(arritem)
+        writer.close()
+
+    else:
+        try:
+            obj = iter(obj)
+        except TypeError:
+            raise TypeError("cannot write {0} to Parquet file".format(type(obj)))
+        try:
+            awkitem = next(obj)
+        except StopIteration:
+            raise ValueError("iterable is empty")
+
+        arritem = convert(awkitem, "cannot write iterator of {0} to Parquet file".format(type(awkitem)))
+        if "schema" not in options:
+            options["schema"] = arritem.schema
+        writer = pyarrow.parquet.ParquetWriter(**options)
+        writer.write_table(arritem)
+
+        try:
+            while True:
+                try:
+                    awkitem = next(obj)
+                except StopIteration:
+                    break
+                else:
+                    writer.write_table(convert(awkitem, "cannot write iterator of {0} to Parquet file".format(type(awkitem))))
+        finally:
+            writer.close()
+
+class _ParquetFile(object):
     def __init__(self, file, cache=None, metadata=None, common_metadata=None):
         self.file = file
         self.cache = cache
@@ -295,7 +495,7 @@ class ParquetFile(object):
         self._init()
 
     def __call__(self, rowgroup, column):
-        return view(self.parquetfile.read_row_group(rowgroup, columns=[column]))[column]
+        return fromarrow(self.parquetfile.read_row_group(rowgroup, columns=[column]))[column]
 
     def tojson(self):
         json.dumps([self.file, self.metadata, self.common_metadata])
@@ -304,10 +504,10 @@ class ParquetFile(object):
     @classmethod
     def fromjson(cls, state):
         return cls(state["file"], cache=None, metadata=state["metadata"], common_metadata=state["common_metadata"])
-        
+
 def fromparquet(file, awkwardlib=None, cache=None, persistvirtual=False, metadata=None, common_metadata=None):
     awkwardlib = awkward.util.awkwardlib(awkwardlib)
-    parquetfile = ParquetFile(file, cache=cache, metadata=metadata, common_metadata=common_metadata)
+    parquetfile = _ParquetFile(file, cache=cache, metadata=metadata, common_metadata=common_metadata)
     columns = parquetfile.type.columns
 
     chunks = []
@@ -315,9 +515,14 @@ def fromparquet(file, awkwardlib=None, cache=None, persistvirtual=False, metadat
     for i in range(parquetfile.parquetfile.num_row_groups):
         numrows = parquetfile.parquetfile.metadata.row_group(i).num_rows
         if numrows > 0:
-            chunk = awkwardlib.Table()
-            for n in columns:
-                chunk[n] = awkwardlib.VirtualArray(parquetfile, (i, n), cache=cache, type=awkwardlib.type.ArrayType(numrows, parquetfile.type[n]), persistvirtual=persistvirtual)
+            if columns == [""]:
+                chunk = awkwardlib.VirtualArray(parquetfile, (i, ""), cache=cache, type=awkwardlib.type.ArrayType(numrows, parquetfile.type[""]), persistvirtual=persistvirtual)
+            else:
+                chunk = awkwardlib.Table()
+                for n in columns:
+                    q = awkwardlib.VirtualArray(parquetfile, (i, n), cache=cache, type=awkwardlib.type.ArrayType(numrows, parquetfile.type[n]), persistvirtual=persistvirtual)
+                    chunk.contents[n] = q
+
             chunks.append(chunk)
             counts.append(numrows)
 
